#!/usr/bin/python

# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import subprocess
import datetime
import time
import os
import signal
import re
import tempfile
import pickle
import shutil
import sys

timeout = 30
first_rev = "e78091ba8904087384ada9a3fc0f19a27e2a4ac0"
verbose_rev = "741f0963bd4431fa41fff945c6c1b3376abcd806"

TIMEOUT = "TIMEOUT"
NOTAPPLICABLE = "NOTAPPLICABLE"
DOESNOTRUN = "DOESNOTRUN"
FAILED = "FAILED"
IGNORE = "IGNORE"
BROKEN = "BROKEN"
DONE = "DONE"
DIFFERENCE = "DIFFERENCE"
NOTRUN = "NOTRUN"
NOTSOLVED = "NOTSOLVED"

#solvers = ["OrderedSolver"]
solvers = ["OrderedSolver", "StackSolver"]

class Stats:
    def run_solver(self, rev, solver, problem):

        try:
            start = datetime.datetime.now()
            args = ["./tests/system_test.py"]
            if self.revisions.index(rev) >= self.revisions.index(verbose_rev):
                args.append("-v")
            args.extend(["-s", solver, problem])
            solvproc = subprocess.Popen(args,
                    stdout=subprocess.PIPE,
                    stderr=file('/dev/null'))

            while solvproc.poll() is None:
                time.sleep(0.2)
                now = datetime.datetime.now()
                if (now - start).seconds> timeout:
                    os.kill(solvproc.pid, signal.SIGKILL)
                    os.waitpid(-1, os.WNOHANG)
                    return TIMEOUT

            output = solvproc.stdout.read()

            if "NOTE: There is no out file" in output:
                return IGNORE

            if "Solver not applicable" in output:
                return NOTAPPLICABLE

            if solvproc.returncode != 0:
                return FAILED

            if "Exception occured" in output:
                return FAILED

            if "Result: Difference!" in output:
                return DIFFERENCE
            
            if "No solution found!" in output:
                return NOTSOLVED
            
            if "Result: OK" in output or "No solution found, as expected" in output:
                rule_app = int(re.compile("^ordinary rule applications:\s+(\d+)$", re.M).search(output).group(1))
                rec      = int(re.compile("^recursion:\s+(\d+)$", re.M).search(output).group(1))
                now = datetime.datetime.now()
                diff = now-start
                ms = diff.seconds * 1000 + diff.microseconds / 1000
                return (ms,rule_app, rec)
            
            #Unparsable output
            print output
        except OSError:
            return DOESNOTRUN

    def check_setup(self):
        try:
            os.environ['PYTHONPATH'] =  'lib/'
            retcode = subprocess.call("./tests/runtests.py",
                    stdout=file('/dev/null','w'),
                    stderr=subprocess.STDOUT)
            if retcode != 0:
                return False
            else:
                return True
        except OSError, e:
            print e
            return False

    def find_testcases(self):
        path_base = 'tests/examples'
        in_filenames = []
        for root, dirs, files in os.walk(path_base):
            in_filenames = [os.path.join(path_base,f) for f in files if f.endswith(".in")]
        return in_filenames

    def find_repo(self):
        return subprocess.Popen(("hg", "root"), stdout=subprocess.PIPE).communicate()[0][:-1]

    def clone_repo(self, dir):
        subprocess.call(("hg", "clone", self.repo, dir))

    def get_revision(self):
        return subprocess.Popen(("hg","tip","--template","{node}\n"), stdout=subprocess.PIPE).communicate()[0][:-1]

    def get_revisions(self):
        self.revisions = subprocess.Popen(("hg","log","--template","{node}\n","--rev",first_rev+":"),
                            stdout=subprocess.PIPE).communicate()[0].rstrip('\n').split('\n')
        self.revision_info = subprocess.Popen(("hg","log","--template",'{author|person}:<br/>{desc|escape|addbreaks}\\0',"--rev",first_rev+":"),
                            stdout=subprocess.PIPE).communicate()[0].rstrip('\0').split('\0')
        self.revision_files = subprocess.Popen(("hg","log","--template",'{files} \\0',"--rev",first_rev+":"),
                            stdout=subprocess.PIPE).communicate()[0].rstrip('\0').split('\0')
        self.revision_name = subprocess.Popen(("hg","log","--template",'{rev}\n',"--rev",first_rev+":"),
                            stdout=subprocess.PIPE).communicate()[0].rstrip('\n').split('\n')
        assert len(self.revisions) == len(self.revision_files), \
            "Not same length: %d (self.revisions) and %d (self.revision_files)" % (len(self.revisions), len(self.revision_files))


    def change_revision(self, rev):
        retcode = subprocess.call(("hg","update","--clean", rev),
                stdout=file('/dev/null'))
        assert retcode == 0, "hg update -r %s failed" % rev

    def measure_revision(self, rev):
        self.change_revision(rev)
        #revision = self.get_revision()
        #assert rev == revision, "%s != %s" % (rev, revision)


        ret = self.check_setup()
        if ret:
            testcases = self.find_testcases()

            for testcase in testcases:
                for solver in solvers:
                    ret = self.run_solver(rev, solver, testcase)
                    if ret:
                        testcasename = os.path.basename(testcase)
                        self.data[(rev,solver, testcasename)] = ret
            
            self.data[(rev,None,None)] = DONE
            print "Revision %s done" % rev
        else:
            self.data[(rev,None,None)] = BROKEN
            print "Revision %s not good" % rev

    def gather_data(self, dir):
        try:
            self.clone_repo(dir)
            os.chdir(dir)
            
            self.get_revisions()

            for revision, files in zip(self.revisions[::-1], self.revision_files[::-1]):
                if not "lib/sevendown" in files:
                    print "Revision %s changed no relevant files" % (revision)
                elif (revision,None,None) in self.data:
                    print "Revision %s already checked" % revision
                else:
                    print "Checking Revision %s" % revision

                    self.measure_revision(revision)
                    self.generate_html()

        except KeyboardInterrupt:
            pass

        self.generate_html()

    def generate_html(self):
        f = file(os.path.join(self.repo, "stats.html"), 'w')
        
        f.write("""
        <html>
        <head>
        <style>
        td { text-align: center}

        .same {
            color:gray;
        }
        .new {
            color:black;
        }
        .tooltip 
            {
                visibility: hidden; 
                position: absolute; 
                top: 0;  
                left: 0; 
                z-index: 2; 

                font: normal 8pt sans-serif; 
                padding: 3px; 
                border: solid 1px;
                background-color: white;
            }
        </style>
        <script type="text/javascript">
        function xstooltip_findPosX(obj) 
        {
          var curleft = 0;
          if (obj.offsetParent) 
          {
            while (obj.offsetParent) 
                {
                    curleft += obj.offsetLeft
                    obj = obj.offsetParent;
                }
            }
            else if (obj.x)
                curleft += obj.x;
            return curleft;
        }


        function xstooltip_findPosY(obj) 
        {
            var curtop = 0;
            if (obj.offsetParent) 
            {
                while (obj.offsetParent) 
                {
                    curtop += obj.offsetTop
                    obj = obj.offsetParent;
                }
            }
            else if (obj.y)
                curtop += obj.y;
            return curtop;
        }
        function xstooltip_show(tooltipId, parentId, posX, posY)
        {
            it = document.getElementById(tooltipId);
            
            if ((it.style.top == '' || it.style.top == 0) 
                && (it.style.left == '' || it.style.left == 0))
            {
                // need to fixate default size (MSIE problem)
                it.style.width = it.offsetWidth + 'px';
                it.style.height = it.offsetHeight + 'px';
                
                img = document.getElementById(parentId); 
            
                // if tooltip is too wide, shift left to be within parent 
                if (posX + it.offsetWidth > img.offsetWidth) posX = img.offsetWidth - it.offsetWidth;
                if (posX < 0 ) posX = 0; 
                
                x = xstooltip_findPosX(img) + posX;
                y = xstooltip_findPosY(img) + posY;
                
                it.style.top = y + 'px';
                it.style.left = x + 'px';
            }
            
            it.style.visibility = 'visible'; 
        }
        function xstooltip_hide(id)
        {
            it = document.getElementById(id); 
            it.style.visibility = 'hidden'; 
        }
        </script>
        <title>7down statistics</title>
        </head>
        <body><h1>7down statistics</h1>
        """)

        f.write("""<p>Letzte Revision: %s (%s):<br/>%s""" % 
                (self.revision_name[-1], 
                self.revisions[-1], 
                self.revision_info[-1]))

        f.write("""
        <p> Legende: ms = <strong>M</strong>illi<strong>s</strong>ekunden, a = Regel-<strong>A</strong>pplikationen, r = <strong>R</strong>ekursionen
        <table border="1">""")
        
        tests = set()

        for (rev, solver, test) in self.data.keys():
            if rev:
                if test:   tests.add(test)

        tests = list(tests)

        def header():
            f.write('<tr><th rowspan="3">Revision</th>')
            for test in tests:
                f.write('<th colspan="%d">%s</th>' % (3*len(solvers), test))
            f.write('</tr>')

            f.write('<tr>')
            for test in tests:
                for solver in solvers:
                    solvername = solver
                    if solver.endswith("Solver"):
                        solvername = solver[:-(len("Solver"))]
                    f.write('<th colspan="%d">%s</th>' % (3, solvername))
            f.write('</tr>')

            f.write('<tr>')
            for test in tests:
                for solver in solvers:
                    f.write('<th>ms</th><th>a</th><th>r</th>')
            f.write('</tr>')

        header()

        oldrev = self.revisions[0]
        for i in range(len(self.revisions)):
            if not "lib/sevendown" in self.revision_files[i]:
                continue

            rev = self.revisions[i]

            name = self.revision_name[i]
            row = ""
            changed = False

            if i == 0:
                changed = True

            row += '<tr>'
            row += '''<td id="td%s"
                onmouseover="xstooltip_show('tt%s', 'td%s', 289, 49);"
                onmouseout="xstooltip_hide('tt%s');">%s</td>''' % (rev, rev, rev, rev, name)

            result = self.data.get((rev,None,None), NOTRUN)
            if i>0:
                old_result = self.data.get((oldrev,None,None), NOTRUN)


            if type(result) == str and result == NOTRUN:
                row += '<td colspan="%d"/>' % (3 * len(solvers) * len(tests))
                if i>0:
                    changed  |=  result != old_result
            elif type(result) == str and result != DONE:
                row += '<td colspan="%d">%s</td>' % (3 * len(solvers) * len(tests), result)
                if i>0:
                    changed  |=  result != old_result
            else:
                for test in tests:
                    for solver in solvers:
                        result = self.data.get((rev,solver,test), NOTRUN)
                        htmlclass = "same"
                        if i>0:
                            old_result = self.data.get((oldrev,solver,test), NOTRUN)
                            cellchanged = result != old_result
                            if cellchanged:
                                htmlclass = "new"

                            if type(result) == str:
                                if result == NOTRUN:
                                    row += '<td class="%s" colspan="3"/>' % (htmlclass)
                                else:
                                    row += '<td class="%s" colspan="3">%s</td>' % (htmlclass, result)
                                changed   |= cellchanged

                            else:
                                if type(old_result) != str:
                                    changed  |= result[1:] != old_result[1:]
                                    for v, o in zip(result,old_result):
                                        if v < o:
                                            color = "green"
                                        elif v > o:
                                            color = "red"
                                        else:
                                            color = "white"
                                        row += '<td style="background-color:%s" class="%s">%d</td>' % (color, htmlclass, v)
                                else:
                                    changed   |= True
                                    for v in result:
                                        row += '<td class="%s">%d</td>' % (htmlclass, v)
                        else:
                            if type(result) == str:
                                row += '<td class="%s" colspan="3">%s</td>' % (htmlclass, result)
                            else:
                                for v in result:
                                    row += '<td class="%s">%d</td>' % (htmlclass, v)

            row += '</tr>'

            if changed:
                f.write(row)
                oldrev = rev

        header()

        f.write('</table>')

        for rev, info, name in zip(self.revisions, self.revision_info, self.revision_name):
            f.write('<div class="tooltip" id="tt%s">%s</div>'% (rev, info))

        f.write('</body></html>')

    def main(self):
        self.repo = self.find_repo()

        datafile = os.path.join(self.repo,"statistics.pickle")

        if os.path.exists(datafile):
            self.data = pickle.load(file(datafile))
        else:
            self.data = {}

        dir = tempfile.mktemp(".7down-stats")
        print "Creating temporary repository in %s"%dir
        try:
            self.gather_data(dir)
            print 'Done. %d cases collected' % len(self.data)

            pickle.dump(self.data, file(datafile,'w'))
        
        finally:
            if os.path.exists(dir):
                print 'Removing temp dir...'
                shutil.rmtree(dir)

if __name__ == '__main__': 
    Stats().main()

# vim:sw=4:ts=4:sts=4:expandtab

