#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


INFINITY = 10**10 + 1
"""
Sentinel value returned by cardinality() when the size is not known. 

Note that this must be larger than any possible number on the field.  At the
moment, all fields can be 10x10 so 10**10+1 suffices.

"""

import sevendown.undo as undo
import sevendown.util as su
import sevendown.config as sc

class Field(object):
    """Fields contain the problem description.
    
    Fields know everything about the problem but for the "block markers".
    """
    
    def __init__(self, height, width):
        """Initialize the field with the given width and height.
        
        After initalization, width and height will be fixed.
        
        """
        # Cells store numbers and digits, numbers stores all numbers.
        self.schema = tuple(tuple([] for j in range(width)) for i in \
            range(height))
        self.cells = []
        self.numbers = []
        self.intermediate_numbers = []
        self.rules = []
        
        # Markers. Not used here, but need to kept around for the GUI
        self.markers = None
        
        # Initialize the Digits matrix.
        self.width = width
        self.height = height
        self.__initDigits(height, width)
    
    def __initDigits(self, height, width):
        """Initializes the internal digits matrix."""
        self.digits = []
        for r in range(height):
            row = []
            for c in range(width):
                digit = Digit(coords= (c, r))
                digit.field = self
                row.append(digit)
            self.digits.append(row)
        
        for t in self.digits:
            for digit in t:
                self.cells.append(digit)
    
    def createNumber(self, name, pos_x, pos_y, length, horizontal):
        """Creates a new Number starting at the position (pos_x, pos_y) and
        with the given length.  The number will be aligned horizontally if
        horizontal is True.
        
        The name of the number will be the value of "name".
        
        Also adds the number to the attributes self.numbers and self.cells.
        
        Returns this Number.
        
        """
        # create the number
        number = FieldNumber(name, pos_x, pos_y, length)
        number.field = self
        
        if horizontal:
            assert pos_x + length <= self.width, "Number at (%d,%d) with length %d does not if in board with width %d" % (pos_x, pos_y, length, self.width)
        else:
            assert pos_y + length <= self.height, "Number at (%d,%d) with length %d does not if in board with height %d" % (pos_x, pos_y, length, self.width)
        
        # initialize the digits
        for i in range(length):
            if horizontal:
                dx, dy = (i, 0)
            else:
                dx, dy = (0, i)
            number.digits[i] = self.digits[pos_y + dy][pos_x + dx]
        
        # add to cells and number
        self.cells.append(number)
        self.numbers.append(number)
        
        return number
    
    def createIntermediateNumber(self, name):
        """Creates a new IntermediateNumber with the given name."""
        number = Number(name)
        number.field = self
        
        self.cells.append(number)
        self.intermediate_numbers.append(number)
        
        return number
    
    def numberForName(self, name):
        """Returns the Number instance for the given name."""
        # only required for loading, being slow is no issue
        numbers = self.numbers + self.intermediate_numbers
        tmp = [n for n in numbers if n.name == name]
        assert len(tmp) > 0, \
            "Could not find the Number for the name %s in %s" % (name,
                str(map (lambda n: n.name, self.numbers)))
        return tmp[0]
    
    def formatResult(self):
        """formats the resulting field, line for line"""
        ret = ""
        for row in self.digits:
            for d in row:
                ret += "%i" % d.value()
            ret += "\n"
        return ret
    
    def registerRule(self, rule):
        """Appends a rule to self.rules and sets rule.field to self."""
        self.rules.append(rule)
        rule.field = self
    
    def getNumbers(self):
        """
        Returns a copy of the Field's list of numbers.
        
        """
        return list(self.numbers)

class Cell(object):
    """
    Super class for Number, Fixed, IntermediateNumber and Digit.
    
    """
    
    def __init__(self):
        self.related_rules = []
    
    def isDone(self): #pragma: no cover
        assert False, "Abstract method isDone() in object of type %s " % \
            type(self).__name__
    
    def isEmpty(self): #pragma: no cover
        assert False, "Abstract method isEmpty() in object of type %s " % \
            type(self).__name__
    
    def cardinality(self): #pragma: no cover
        """
        Return an upper limit for the possible values of the cell.
        
        Return value is in [0, INFINITY).
        
        """
        assert False, "Abstract cardinality() in object of type %s " % \
            type(self).__name__
    
    def isSplittable(self):
        """returns Boolean, when the set can be partitioned"""
        return False
    
    def split(self): #pragma: no cover
        """returns a list of somthing that can be passed to force"""
        assert False, "Abstract method split() in object of type %s " % \
            type(self).__name__
    
    def invalidate(self, i): #pragma: no cover
        """
        Removes i from the feasible values.  Returns True iff i was in the
        set of feasible values, returns False otherwise.
        
        Abstract, override in your class.
        
        """
        assert False, "Abstract method invalidate() in object of type %s " % \
            type(self).__name__
    
    def intersect(self, alsoValid):
        """
        Intersects the list of feasible digits with the given set.   Returns
        True if the Number was changed.
        
        Abstract, override in your class.
        
        """
        assert False, "Abstract method intersect() in object of type %s " % \
            type(self).__name__
        
    def force(self, values): #pragma: no cover
        """force-set the value to something returned by split()"""
        assert False, "Abstract method force() in object of type %s " % \
            type(self).__name__
    
    def setFeasibleValues(self, values): #pragma: no cover
        """
        Set new feasible values, has no undoable semantics.
        
        Abstract, override in your class.
        
        """
        assert False, "Abstract setFeasibleValues() in object of type %s " % \
            type(self).__name__
    
    def getFeasibleValues(self): #pragma: no cover
        """
        Returns a copy of the set of feasible values for this digit.
        
        Abstract, override in your class.
        """
        assert False, "Abstract getFeasibleValues() in object of type %s " % \
            type(self).__name__
    
    def value(self): #pragma: no cover
        """returns the value. Makes only sense if isDone == True"""
        assert False, "Abstract method value() in object of type %s " % \
            type(self).__name__
    
    def getRelatedRules(self):
        """returns the list of rules that subscribed to this cell"""
        return self.related_rules
    
    def addToRelatedRules(self, rule):
        """adds the ruel to the list ob subscribed cells"""
        self.related_rules.append(rule)
    
    def __hash__(self):
        return hash(self.name)


class Digit(Cell):
    """
    A Digit represents a set of values that are still possible for a digit
    of a number.  It can represent a square on a field or be anonymous as
    the digit of an intermediate or fixed number.
    
    """
    
    def __init__(self, value=None, coords=None):
        """
        Initialize the Digit.
        
        Iff value is given then the valid values for the Digit are set to the
        set containing only value.  Iff name is given then the Digit's name
        is set to name.
        
        """
        Cell.__init__(self)

        self.coords = coords
        if coords:
            self.name = "(%d,%d)" % self.coords
        else:
            self.name = "(?,?)"
        
        # The underscore in front of _feasible_values is there to indicate
        # that the variable must not be accessed from the outside!
        if value is None:
            self._feasible_values = set(range(0, 10))
        else:
            self._feasible_values = set([value])
        
        self.field = None
    
    def __str__(self):
        """
        Returns string representation of the Digit.
        
        Will be have the format "<X,Y>".
        
        """
        return "<%s>" % self.name

        
    def isDone(self):
        """Returns True iff there is exactly one feasible value left."""
        return len(self._feasible_values) == 1
    
    def isEmpty(self):
        """Returns True iff the no value is feasible any more."""
        return len(self._feasible_values) == 0    
    
    def cardinality(self):
        """
        Returns the number of feasible values for this Digit.
        
        Return value will be within [0, 10].
        
        """
        return len(self._feasible_values)
    
    def isSplittable(self):
        """Returns True iff the Digit has more than one feasible value."""
        return len(self._feasible_values) > 1
    
    def split(self):
        """
        Return a list of sets in which the Digit's feasible value are split.  
        
        """
        if sc.DIGIT_SPLIT == sc.DIGIT_SPLIT_HALF:
            # perform a 50-50 split
            s1 = self._feasible_values.copy()
            s2 = set()
            for i in range(0,len(s1)/2):
                s2.add(s1.pop())
            return [s1, s2]
        elif sc.DIGIT_SPLIT == sc.DIGIT_SPLIT_SINGLETON:
            # perform a singleton split
            return [set((n,)) for n in self._feasible_values ]
        
        assert False, "Invalid value for sc.DIGIT_SPLIT!" #pragma: no cover
    
    @undo.undoable('_feasible_values')
    def invalidate(self, i):
        """
        Removes i from the feasible values.  Returns True iff i was in the
        set of feasible values, returns False otherwise.
        
        """
        assert isinstance(i, int) or isinstance(i, long)
        
        if not i in self._feasible_values:
            return False
        
        # actually remove i from self._feasible_values
        self._feasible_values.remove(i)
        return True
    
    @undo.undoable('_feasible_values')
    def intersect(self, alsoValid):
        """
        Intersects the list of feasible digits with the given set.
        
        Returns True if the Number was changed.
        """
        oldcard = len(self._feasible_values)
        self._feasible_values.intersection_update(alsoValid)
        return oldcard != len(self._feasible_values)
    
    @undo.undoable('_feasible_values')
    def force(self, newValues):
        """
        Force-sets the feasible values of a Digit.
        
        This method is undoable.
        
        """
        assert type(newValues) == set, "Digit.force() expects a set()"
        
        if newValues != self._feasible_values:
            self._feasible_values = newValues
            return True
        else:
            return False
    
    def setFeasibleValues(self, values):
        """
        Sets values as the new set of feasible values.
        
        If the values are set using this method then they are not undoable.
        
        """
        self._feasible_values = values
    
    def getFeasibleValues(self):
        """
        Returns the set of feasible values for this digit.

        Do not modify this set. If you want, make a copy!
        
        """
        return self._feasible_values
    
    def value(self):
        """
        Returns the single feasible value of this Digit.
        
        Can only be called when the Digit isDone(), i.e. the set of feasible
        values has a cardinality of one.
        
        """
        assert self.isDone()
        return list(self._feasible_values)[0]
    

class SetTooLargeException(Exception):
    """Thrown when a to be returned set is too large."""


class Number(Cell):
    def __init__(self, name, length=None):
        """
        Creates a number representation.
        
        Lengths might be unknown (e.g. for an intermediate Number).
        
        Note that this does *not* create any Digits for the number.  It only
        initializes the list self.digits.  The digits have to be set by the
        Field.
        
        """
        Cell.__init__(self)
        self.name = name
        
        self.length = length
        
        # initialize the digits
        if length:
            self.digits = [None] * length
        else:
            self.digits = None
        
        # initialize rules
        self.rules = []
        
        # a reference to the Field object this number is on
        self.field = None
        
        # list of still possible values. None means: look at digits
        self._feasible_values = None
    
    def __str__(self):
        """Returns a human readable version of the number."""
        # TODO: Also print possible digits?s
        return self.name
    
    def _isDigitDone(self):
        """
        Returns True iff all digits are "isDone()".
        
        Do not call directly, used by isDone().
        
        """
        if not self.hasDigits():
            return False
        
        for digit in self.digits:
            if not digit.isDone():
                return False
        return True
    
    def isDone(self):
        """
        Returns True iff the Number has only one feasible value left.
        
        This is the case if all Digits are "isDone()" or the cardinality of
        the set of explicit feasible values is one.
        
        """
        if self._feasible_values == None:
            return self._isDigitDone()
        else:
            return len(self.getFeasibleValues()) == 1
    
    def _isDigitEmpty(self):
        """
        Returns True iff all digits are "isEmpty()".
        
        Do not call directly, used by isEmpty().
        
        """
        if self.hasDigits():
            for digit in self.digits:
                if digit.isEmpty():
                    return True
        return False
    
    def isEmpty(self):
        """
        Returns True iff the Number is empty, no values are feasible.
        
        """
        if self._feasible_values == None:
            return self._isDigitEmpty()
        else:
            return len(self._feasible_values) == 0
    
    def cardinality(self):
        """
        Returns a number that is not smaller than the possible values returned
        by getFeasibleValues().
        
        This is either the number of possible numbers that can be generated
        from the digits, the size of the values set explicitely or INFINITY.
        
        """
        if self._feasible_values == None:
            if self.digits == None:
                return INFINITY
            else:
                return self._digitCardinality()
        else:
            return len(self._feasible_values)
    
    def _getFeasibleDigitValues(self):
        """
        Returns an iterator of the possible values for this number.
        
        Do not call directly, used by getFeasibleValues().
        
        """
        digit_values = [digit.getFeasibleValues() for digit in self.digits]
        return su.to_int(su.xcombine(digit_values))
    
    def getFeasibleValues(self):
        """
        Returns a set of all valid number.  Automatically generates the set
        from the digits' possible values.
        
        Raises a SetTooLargeException if the generated set would be too large.
        
        """
        if self.cardinality() > sc.MAX_EXPLICIT_NUMBERS:
            s = "The set to be returned would be too large: %d." % \
                self.cardinality()
            raise SetTooLargeException(s)
        
        if self._feasible_values == None:
            self.force(set(self._getFeasibleDigitValues()))
        
        return self._feasible_values
    
    def setFeasibleValues(self, values):
        """
        Sets the given set of values as the set of feasible values.
        
        Calls to this method are not undoable.
        
        """
        if self.length:
            max = 10**self.length -1
            pred = lambda v: v <= max
        else:
            pred = lambda v: True
        self._feasible_values = set(v for v in values if pred(v))
    
    @undo.undoable('_feasible_values')
    def force(self, newValues):
        """
        Forces the list of valid numbers to the newValues.
        
        Calls to this method are undoable.  Returns True iff the Number was
        changed.
        
        """
        assert type(newValues) == set, "Number.force() expects a set"
        
        assert len(newValues) <= sc.MAX_EXPLICIT_NUMBERS, \
            "Too many numbers passed to force: %d" % len(newValues)
        
        if newValues != self._feasible_values:
            self.setFeasibleValues(newValues)
            return True
        else:
            return False
    
    @undo.undoable('_feasible_values')
    def invalidate(self, value):
        """
        Removes this value from the list of possible value, if present.
        
        Calls to this method are undoable. Returns True if the Number was
        changed.
        
        """
        if self._feasible_values != None and value in self._feasible_values:
            self._feasible_values.discard(value)
            return True
        else:
            return False
    
    @undo.undoable('_feasible_values')
    def intersect(self, alsoValid):
        """
        Intersects the list of valid numbers with the given set.
        
        Calls to this method are undoable.  Returns True if the Number was
        changed.
        
        """
        assert len(alsoValid) <= sc.MAX_EXPLICIT_NUMBERS, str(len(alsoValid))
        if self._feasible_values == None:
            return self.force(alsoValid)
        else:
            oldcard = len(self._feasible_values)
            self._feasible_values.intersection_update(alsoValid)
            
            return oldcard != len(self._feasible_values)
    
    @undo.undoable('_feasible_values')
    def invalidateSet(self, invalid_set):
        """
        Invalidate the set "invalid_set".
        
        If we don’t have an explicit set set, there is nothing we can do.  
        Returns True iff the Number was changed.
        
        """
        if self._feasible_values == None:
            return False
        else:
            oldcard = len(self._feasible_values)
            self._feasible_values.difference_update(invalid_set)
            
            return oldcard != len(self._feasible_values)
    
    def value(self):
        assert self.isDone()
        return list(self.getFeasibleValues())[0]
    
    def _digitCardinality(self):
        """
        Returns the cardinality of the possible values from the digits.
        
        Do not call directly, used by cardinality().
        
        """
        if self.hasDigits():
            p = 1
            for digit in self.digits:
                p *= digit.cardinality()
            return p
        return sc.ALMOST_INFINITE
    
    def minimum(self):
        """
        Return smallest possible integer that could be represented by the
        Numbers' digits.
        
        """
        # TODO: What happens if self.cardinality() == 0?
        return self.minMaxHelper(min)
    
    def maximum(self):
        """
        Return largest possible integer that could be represented by the
        Numbers' digits.
        
        """
        # TODO: What happens if self.cardinality() == 0?
        return self.minMaxHelper(max)
    
    def minMaxHelper(self, func):
        if self._feasible_values != None:
            res = None
            for val in self._feasible_values:
                if res:
                    res = func(res, val)
                else:
                    res = val
            return res
        else:
            res = 0
            for digit in self.digits:
                res = (res * 10) + func(digit.getFeasibleValues())
            return res
    
    def hasDigits(self):
        """Returns true if this Number instance knows about its digits"""
        if self.digits:
            return self.digits[0] != None
        else:
            return False
    

class FieldNumber(Number):
    """
    Conveniece class.
    
    Use this to create Number objects that appear on a Field.  The Number's
    name, position and length attributes will be set correctly.
    
    """
    
    def __init__(self, name, pos_x, pos_y, length):
        """Creates a number representation
        
        
        Note that this does *not* create any Digits for the number.  It only
        initializes the list self.digits.  The digits have to be set by the
        Field.
        """
        
        Number.__init__(self, name, length = length)
        
        self.pos = (pos_x, pos_y)
    

class Fixed(Number):
    """
    Convenience class that allows to initialize a number with all digits fixed
    to singleton sets easily.
    
    """
    
    def __init__(self, value):
        Number.__init__(self, str(value), len(str(value)))
        
        value_digits = su.digits(value)
        for i in range(len(value_digits)):
            self.digits[i] = Digit(value=value_digits[i])
        
        self._feasible_values = set([value])
    
    def value(self):
        """
        Returns the fixed's value.
        
        """
        assert len(self._feasible_values) == 1
        return list(self._feasible_values)[0]

# vim:sw=4:ts=4:sts=4:expandtab
