#! /usr/bin/env python
# -*- coding: utf-8 -*- 
# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import copy

import sevendown.graph as sg
import sevendown.loader as sl
import sevendown.rules as sr

class CellPrinter(object):
    """Gives a nice one-line-representation of a cell"""
    def __init__(self, cell):
        self.cell = cell

    def run(self):
        if self.cell._feasible_values == None:
            vals = "(unknown)"
        elif len(self.cell._feasible_values)<=10:
            vals = "(%s)" % (str(list(self.cell._feasible_values)))
        else:
            vals = "(%s... %d more)" % (str(list(self.cell._feasible_values)[:10]), len(self.cell._feasible_values)-10)
        if self.cell.hasDigits():
            digits = "(d)"
        else:
            digits = "( )"
        return "%s %s %s\n" % (str(self.cell), digits, vals)

class FieldPrinter(object):
    """This class allows to convert a field into a pretty string 
    representation.
    
    """
    
    # +-------+
    # |A      |
    # |       |
    # | 01234 |
    # | 56789 |
    # +-------+
    
    
    def __init__(self, field):
        self.screen = None # the string to print the field on
        self.field = field
    
    def printBorders(self):
        """Prints the borders of the field on the given string screen."""
        # print horizontal border
        for y in [0, self.field.height * 6]:
            for x in range(self.field.width * 7 + 1):
                self.screen[y][x] = '-'
        
        # print vertical border
        for y in range(1, len(self.screen) - 1):
            for x in [0, len(self.screen[y]) - 1]:
                self.screen[y][x] = '|'
        
        # print +
        for y in range(self.field.height + 1):
            for x in range(self.field.width + 1):
                self.screen[y*6][x*7] = '+'
    
    def printMarkers(self):
        """Prints the markers on the screen."""
        for (n,x,y) in self.field.markers.values():
            self.screen[y*6+1][x*7+1] = n
    
    def printDigitsPossibleValues(self):
        for y, line in enumerate(self.field.digits):
            for x, digit in enumerate(line):
                for i in [0, 1]:
                    for j in range(5):
                        if i * 5 + j in digit.getFeasibleValues():
                            c = str(i * 5 + j)
                            self.screen[y * 6 + 3 + i][x * 7 + 1 + j] = c
    
    def linewidth(self):
        return self.field.width * 7 + 1
    
    def formatRules(self):
        ret = ""
        ret += "Rules\n"
        ret += "-----\n"
        ret += "\n"
        ret += "\n".join([str(r) for r in self.field.rules])
        ret += "\n"
        return ret
    
    def printRules(self):
        print self.formatRules
    
    def run(self):
        # Create string to write the field on later.
        empty_line = [" "] * (self.linewidth())
        lines = [copy.copy(empty_line) \
            for i in range(self.field.height * 6 + 1)]
        self.screen = lines
        
        self.printBorders()
        self.printMarkers()
        self.printDigitsPossibleValues()
        
        field = "\n".join(["".join(line) for line in self.screen])

        intermediates = ""
        for cell in self.field.cells:
            if isinstance(cell, sg.Number):
                intermediates += CellPrinter(cell).run()
        #rules = "\n".join([str(r) for r in self.field.rules])
        
        return '''
%s

Numbers
--------------
%s
''' % (field, intermediates)
    

def printField(field):
    """Takes a Field() and prints it to the current terminal.
    """
    print FieldPrinter(field).run()


def formatField(field):
    """Takes a Field() and returns a hopefully nice representation as a 
    string"""

    # Brainstorming: 
    # +-+
    # |A| 
    # + +
    #

    r = ""
    for row in range(field.height):
        if row == 0:
            for col in range(field.width):
                if col == 0:
                    r += "+"
                r += "-"
                r += "+"
            r += "\n"

        for col in range(field.width):
            if col == 0:
                r += "|"

            # This is the real content:
            c = " "
            for (n,x,y) in field.markers.values():
                if x == col and y == row:
                    c = n
            r += c

            if sl.BORDER_RIGHT in field.schema[row][col]:
                r += "|" 
            else:
                r +=  " "
        r += "\n"

        for col in range(field.width):
            if col == 0:
                r += "+"

            if sl.BORDER_BOTTOM in field.schema[row][col]:
                r += "-" 
            else:
                r +=  " "

            r += "+"

        r += "\n"

    r += "\n"

    r += "Rules:\n"
    for rule in field.rules:
        if not isinstance(rule, sr.NoLeadingZeroRule):
            r += str(rule) 
            r += "\n"

    return r

# vim:sw=4:ts=4:sts=4:expandtab
