# -*- coding: utf-8 -*- 

# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""This file defines all Rule classes."""

import math
import random
import operator as op
from itertools import takewhile, dropwhile

import sevendown
import sevendown.graph as sg
import sevendown.config as sc
import sevendown.util as su
from sevendown.binlist import BinList 

# small priority => high priority

PRIORITY_DIGIT_RULES = 0
"""Priority for rules on Digits."""
PRIORITY_ONCE_RULES = -100
"""Priority for to-be-applied once rules."""
PRIORITY_ONCE_RULES_DONE = 10000
"""Priority for to-be-applied rules that have already been applied."""

PRIORITY_UNARY_RULE = -10
PRIORITY_BINARY_RULE = 0
PRIORITY_OPERATOR_RULE = 10
PRIORITY_UNIQUE_RULE = -11

PRIORITY_NUMBER_RULE = 100

def createRule(name, *params):
    """Creates a Rule object with the given name and the given parameters."""
    assert NAME_TO_RULE.has_key(name)
    
    return NAME_TO_RULE[name](*params)

def createOperatorRule(name, *params):
    """Creates an OperatorRule object with the given name and the given parameters."""
    assert OPERATOR_TO_RULE.has_key(name)
    
    return OPERATOR_TO_RULE[name](*params)


class Rule(object):
    """
    Implements the Rules from the problem description.
    
    """
    
    def __init__(self):
        self.field = None
        """The field this rule belongs to."""
        
        self.priority = 0
        """The rule's priority."""
    
    def apply(self, causing_cells=None): #pragma: no cover
        """
        This is where the big show is happening: The rule will look into the
        cells and try to reduce the possible values of the cell.
        
        causing_cells, if set, contains the set of cells that were modified since the last
        call to apply().
        
        Returns a list of the cells that the rule has modified.
        
        Override in child classes.
        
        """
        assert False, "Abstract method apply() in object of type %s " % \
            type(self).__name__
    
    def register(self): #pragma: no cover
        """
        Registers the rule with all of its arguments.
        
        Override in child class.
        
        """
        assert False, "Abstract method register() in object of type %s " % \
            type(self).__name__
    
    def normalize(self): #pragma: no cover
        """
        Normalizes the Rule.
        
        Returns a list of rule to replace the current rule if the rule can
        be replaced by simpler rules.  Can possibly include the rule itself if 
        the rule wants to add more rules.
        
        If no normalization is required, return [self]. This is also the 
        default implementation.
        
        Rule object can rely that this method is called at least once, before 
        the Solver is run, but it might be called more than once.
        
        """
        return [self]
    
    def __le__(self, other):
        """Rules are ordered by their priority."""
        assert isinstance(other, Rule)
        return self.priority <= other.priority
    
    def __hash__(self):
        return hash(self.name)
    
    def getPriority(self):
        """
        Returns the integer attribute "priority" of the Rule.
        
        A Rule with a higher priority is applied earlier.  This method returns
        self.priority.  Override in your subclass to get dynamic priority
        determination.
        
        """
        return self.priority
    
    def applyPredicate(self, cell, pred):
        """
        Helper function: If the cell is of low enough cardinality, remove
        those value that do not match the predicate
        
        Returns True if the cell was changed"""

        if cell.cardinality() < sc.MAX_EXPLICIT_NUMBERS:
            correct_values = filter(pred, cell.getFeasibleValues())
            return cell.intersect(correct_values)
        return False
    
    def detailedDescription(self):
        """
        Called by the GUI. Should return a rather verbose description
        of what the rule is about, using full sentences and using the
        proper names for the related cells.
        """
        return "Undocumented Rule"
    
    def relatedCells(self):
        """
        Returns the list of cells that have something to do with
        this rule.
        """
        return []
    

class BinaryRule(Rule):
    """
    Parent class for all classes with two parameters.
    
    The parameters are stored in self.arg1 and self.arg2.  Override this
    class to implement your own binary rules.
    
    """
    
    def __init__(self, arg1, arg2):
        """Initialize the rule with the two given arguments."""
        Rule.__init__(self)
        self.name = 'UNKNOWN'
        self.arg1 = arg1
        self.arg2 = arg2
        self.register()
        
        self.priority = PRIORITY_BINARY_RULE
    
    def register(self):
        """Register with dependencies. Override in subclass if necessary."""
        self.arg1.addToRelatedRules(self)
        self.arg2.addToRelatedRules(self)
    
    def __str__(self):
        """Returns a human readable string representation of the rule."""
        return "%s(%s, %s)" % (self.name, self.arg1, self.arg2)
    
    def relatedCells(self):
        return (self.arg1, self.arg2)
    

class UnaryRule(Rule):
    """A rule with one parameter."""
    
    def __init__(self, arg):
        """Initialize the unary rule object."""
        Rule.__init__(self)
        self.name = "UNKNOWN"
        self.arg = arg
        self.register()
        
        self.priority = PRIORITY_UNARY_RULE
    
    def register(self):
        """Register with dependencies. Override in subclass if necessary."""
        self.arg.addToRelatedRules(self)
    
    def __str__(self):
        """Returns a human readable string representation of the rule."""
        return "%s(%s)" % (self.name, self.arg)
    
    def relatedCells(self):
        return (self.arg,)
    

class DescendingRule(UnaryRule):
    """
    A rule on a number whose values must be descending.
    
    Here, "descending" is defined as "nonascending" so "111" also qualifies
    as a "descending" number.
    
    """
    
    def __init__(self, arg):
        assert isinstance(arg, sg.Number), "only works on Numbers"
        
        UnaryRule.__init__(self, arg)
        self.name = 'AB'

        self.priority = PRIORITY_UNIQUE_RULE
    
    def apply(self, causing_cells=None):
        """
        For each of the numbers digits, determine the largest possible value
        K of this digit (must not be larger as any value left of it).  
        Remove all values greater or equal to K.
        
        Performs the analogue task for "smallest possible value for each 
        digit".
        
        """
        
        result = set()
        
        if self.arg.hasDigits():
            # parameters for the elimination loop below
            params = [(10, max, op.le, su.identity), (-1, min, op.ge, reversed)]
            
            # eliminate impossible values
            for initial_limit, minmax, comp, maybe_reverse in params:
                limit = initial_limit
                for digit in maybe_reverse(self.arg.digits):
                    valid_ones = [v for v in digit.getFeasibleValues() \
                        if comp(v, limit)]
                    
                    if len(valid_ones) == 0:
                        limit = initial_limit
                    else:
                        limit = minmax(valid_ones)
                    
                    if digit.force(set(valid_ones)):
                        result.add(digit)

        if self.arg.cardinality() < sc.MAX_EXPLICIT_NUMBERS:
            # Also look through the explicit numbers, whether we want
            # to change something
            def check(n):
                d = su.digits(n)
                return reduce(op.and_, map(op.ge, d, d[1:]), True)
            if self.applyPredicate(self.arg, check):
                result.add(self.arg)
            
        return list(result)

    def detailedDescription(self):
        return "The digits of %s are sorted in descending order." % (str(self.arg))

    def register(self):
        """
        Register with all digits of the underlying number and with the number
        itself.
        
        """
        UnaryRule.register(self)
        if isinstance(self.arg, sg.Number) and self.arg.hasDigits():
            for digit in self.arg.digits:
                digit.addToRelatedRules(self)
    

class NoLeadingZeroRule(UnaryRule):
    """
    The argument must be a number, the value 0 is removed from the first 
    digit.
    
    """
    
    def __init__(self, arg):
        assert isinstance(arg, sg.Number), "only works on Numbers"
        
        UnaryRule.__init__(self, arg)
        self.name = "NLZ"
        self.applied = False
        
        # This rule is only to be applied in the beginning of the solution.
        # Thus, it gets a higher priority.
        self.priority = PRIORITY_ONCE_RULES
    
    def register(self):
        """Overwriting the method so no registering takes place."""
        pass
    
    def apply(self, causing_cells=None):
        """
        Remove 0 from the first digit's valid values, return an list
        containing all changed digits.
        
        """
        self.priority = PRIORITY_ONCE_RULES_DONE
        if self.arg.digits[0].invalidate(0):
            return [self.arg.digits[0]]
        else:
            return []
     
    def detailedDescription(self):
        return "The first digit of %s is not zero." % (str(self.arg))


class PalindromeRule(UnaryRule):
    """
    Unary rule on a Number that asserts that the argument is a Palindrome.
    
    This Rule only supports the normalize-operation, as it can be replaced by
    simple EqualityRule objects.
    
    """
    
    def __init__(self, arg):
        assert isinstance(arg, sg.Number), "argument must be a Number"
        
        UnaryRule.__init__(self, arg)
        self.name = 'PAL'
    
    def register(self):
        """Only register if argument has no digits."""
        if not self.arg.hasDigits():
            UnaryRule.register(self)
    
    def normalize(self):
        """
        Create EqualityRule objects on the argument's Digit objects that
        enforce the argument being a palindrome.  Return a list of all new
        EqualityRules.
        
        """
        new_rules = []
        
        if self.arg.hasDigits():
            for i in range(len(self.arg.digits) / 2):
                rule = EqualityRule(self.arg.digits[i],
                    self.arg.digits[-(i+1)])
                new_rules.append(rule)
        
        return new_rules
    
    def apply(self, causing_cells=[]):
        """
        Only called when registered and thus only when underlying Number
        has no digits.
        
        """
        def is_palindrome(x):
            """Returns True iff x is a palindrome integer."""
            digits = su.digits(x)
            k = len(digits)
            for i in range(k/2):
                if digits[i] != digits[k - i - 1]:
                    return False
            return True
        
        if self.arg in causing_cells and \
                self.applyPredicate(self.arg, is_palindrome):
            return [self.arg]
        else:
            return []
    
    def detailedDescription(self):
        return "The number %s is an palindrome." % (str(self.arg))
    

class RemoveValuesFromDigitsRule(UnaryRule):
    """
    Parameter must be a Digit, removes certain values from the _feasible_values
    set of the Digit.  This is used by the PrimeRule on the last digit in
    normalize().
    
    """
    
    def __init__(self, arg, in_feasible_values):
        assert isinstance(arg, sg.Digit), "argument must be a Digit"
        
        UnaryRule.__init__(self, arg)
        self.in_feasible_values = in_feasible_values
        self.name = 'REMOVE'
        
        # This rule is only to be applied in the beginning of the solution.
        # Thus, it gets a higher priority.
        self.priority = PRIORITY_ONCE_RULES
    
    def __str__(self):
        """String conversion."""
        return "%s[%s](%s)" % (self.name, 
            su.iterable2string(self.in_feasible_values), str(self.arg))
    
    def register(self):
        """Overwrite so nothing is registered."""
        pass
    
    def apply(self, causing_cells=None):
        """
        Remove all in_feasible_values from the _feasible_values of the underlying
        Digit.
        
        """
        self.priority = PRIORITY_ONCE_RULES_DONE
        changed = False
        for i in self.in_feasible_values:
            changed |= self.arg.invalidate(i)
        
        if changed:
            return [self.arg]
        else:
            return []
       
    def detailedDescription(self):
        return "Removes the digits %s as possible values from the digit at %s" % (", ".join(map(str,self.in_feasible_values)),str(self.arg))
  

class PrimeRule(UnaryRule):
    """
    Unary rule on a Number that asserts that the argument is a prime number.
    
    TODO: This currently relies on a list of prime numbers.
    TODO: This only works up the largest prime number in the list.
    TODO: Implement Rabin-Miller test for larger primes.
    
    """
    primelist = None
    primecache = {}
    
    def __init__(self, arg):
        assert isinstance(arg, sg.Number), "argument must be a number"
        
        UnaryRule.__init__(self, arg)
        self.name = 'PZ'
        if PrimeRule.primelist is None:
            PrimeRule.primelist = BinList(sevendown.__path__[0] + "/../../data/primenumbers.dump")
        
    def normalize(self):
        """
        If the argument is a Number then return self and a new 
        RemoveValuesFromDigitsRule that removes the values 0,2,4,5,6,8 from
        the last digit of the argument.
        
        """
        if self.arg.hasDigits():
            digit = self.arg.digits[-1]
            in_feasible_values = [0,2,4,5,6,8]
            return [self, RemoveValuesFromDigitsRule(digit, in_feasible_values)]
        else:
            return [self]
    
    def apply(self, causing_cells=None):
        """
        If the underlying number isDone() then its value will be checked to
        be a prime.
        
        TODO: Set number's possible values to the primes its digits allow.
        
        """
        changed = False

        changed |= self.applyPredicate(self.arg, self.is_prime)

        if self.arg.length == 4 and self.arg.cardinality() > sc.MAX_EXPLICIT_NUMBERS:
            primes = set(
                    takewhile(lambda n: n<10**4,
                        dropwhile(lambda n: n < 10**3,
                            PrimeRule.primelist)))
            changed |= self.arg.intersect(primes)
        
        if changed:
            return [self.arg]
        else:
            return []

    def rabin_miller(self, n):
        d = n-1
        while d & 1 == 0: 
            d >>= 1
        for k in range(1,20):
            a = random.randint(1,n-1)
            t = d
            y = pow(a,t,n)
            while (t != n-1) & (y != 1) & (y != n-1):
              y = (y * y) % n
              t <<= 1
            if (y != n-1) & (t & 1 == 0):
                return False
        return True

    
    def is_prime(self, n):
        """
        Returns True if n is prime.
        
        This is currently done by checking whether the value is in a list
        of prime numbers (including all prime numbers up to 10^7).
        
        Actually Rabin-Miller test with k=20 for numbers with length greater than 8.
        
        """
        if n <= PrimeRule.primelist[-1]:
            return n in PrimeRule.primelist
        else: 
            'Miller-Rabin primality test'
            if n == 2: 
                return True 
            if (n == 1) | (n & 1 == 0): 
                return False

            if n not in PrimeRule.primecache:
                PrimeRule.primecache[n] = self.rabin_miller(n)

            return PrimeRule.primecache[n]
                
    def detailedDescription(self):
        return "The number %s is a prime." % (str(self.arg))


class PowerRule(UnaryRule):
    """
    Unary rule on Number objects that asserts that the argument is a power
    of an integer.
    
    This class is extended by SquareRule and CubicRule.  Do not instanciate
    directly!
    
    """
    
    def __init__(self, arg, exp):
        assert isinstance(arg, sg.Number), "argument must be a Number"
        
        UnaryRule.__init__(self, arg)
        self.exp = exp
    
    def apply(self, causing_cells=None):
        """
        Reduces set of numbers if cardinality is smaller than 
        sc.SQUARE_EXHAUSTIVE_LIMIT or length of numbers is smaller than
        SQURE_LENGTH_LIMIT, else returns an empty set.
        
        """
        number = self.arg
        changed = False
        numlen = number.length
        

        if number.cardinality() < sc.MAX_EXPLICIT_NUMBERS:
            def check(n):
                return n == round(math.pow(n, 1.0/float(self.exp))) ** self.exp
            changed |= self.applyPredicate(self.arg, check)

        elif numlen and numlen < sc.POTENCE_LENGTH_LIMIT * self.exp:
            # if the number is short enough (i.e. it has few enough digits)
            # then we can simply generate all possible numbers.
            # TODO: 10^{k-1} and 10^{k} are pretty rough estimates
            powers = su.pot_numbers(self.exp, math.pow(10, numlen - 1),
                math.pow(10, numlen))
            ret = number.intersect(set(powers))
            if ret: 
                changed = True
        
        if changed:
            return [number]
        else:
            return []
    

class SquareRule(PowerRule):
    """Specialization of PowerRule with the exponent set to 2."""
    
    def __init__(self, arg):
        PowerRule.__init__(self, arg, 2)
        self.name = 'Q'
      
    def detailedDescription(self):
        return "The number %s is a squared integer." % (str(self.arg))
   

class CubicRule(PowerRule):
    """Specialization of PowerRule with the exponent set to 3."""
    
    def __init__(self, arg):
        PowerRule.__init__(self, arg, 3)
        self.name = 'K'
         
    def detailedDescription(self):
        return "The number %s is a cubed integer." % (str(self.arg))


class UniqueDigitsRule(UnaryRule):
    """
    Unary rule on Number objects, asserts digits' values must be unique.
    
    """
    
    def __init__(self, arg):
        assert isinstance(arg, sg.Number), "argument must be a Number"
        UnaryRule.__init__(self, arg)
        self.name = 'U'
    
    def register(self):
        """
        Register with all digits of the underlying number.
        
        """
        for digit in self.arg.digits:
            digit.addToRelatedRules(self)
    
    def apply(self, causing_cells=None):
        """
        Apply the "unique digits" rule:
        
        * if there is a digit with only one valid value left then remove
          it from all other values
        
        """
        # the digits that are done
        if causing_cells is None:
            causing_cells = self.arg.digits

        # check if there are actually enough digits left
        occuring_digits = set()
        for d in self.arg.digits:
            occuring_digits |= d.getFeasibleValues()
        if len(occuring_digits) < self.arg.length:
            # not enugh digits left
            self.arg.digits[0].force(set())
            return [self.arg.digits[0]]

        # the digits that have been changed
        result = set()

        todo_digits = set(d for d in self.arg.digits if d.isDone() and d in causing_cells)
        
        while todo_digits:
            done_digit = todo_digits.pop()
            if done_digit.isEmpty():
                # done_digit is empty, must have been added to result in for
                # loop below, breaking here will return the empty digit which
                # will break recursion
                assert done_digit in result
                break
            
            for digit in self.arg.digits:
                # skip current
                if digit == done_digit:
                    continue
                
                # remove valid value from digit otherwise
                if done_digit.value() in digit.getFeasibleValues():
                    result.add(digit)
                    digit.invalidate(done_digit.value())
                    
                    if digit.isDone():
                        todo_digits.add(digit)
                    if digit.isEmpty():
                        # no more values, digit is in result, return will
                        # break recursion
                        break

        def check(n):
            return len(set(su.digits(n))) == len(su.digits(n))
        if self.applyPredicate(self.arg, check):
            result.add(self.arg)

        return list(result)
         
    def detailedDescription(self):
        return "The number %s has unique digits." % (str(self.arg))


class CrossSumRule(BinaryRule):
    """
    Binary rule, the cross sum of the second argument must be equal to the
    first argument's value.
    
    """
    
    def __init__(self, value, number):
        """Initialize the rule with the given cross sum value and number."""
        BinaryRule.__init__(self, value, number)
        self.name = "CrossSumIs"
    
    def register(self):
        """Register itself with arguments and their digits (if any)."""
        BinaryRule.register(self)
        
        for arg in [self.arg1, self.arg2]:
            if arg.hasDigits():
                for d in arg.digits:
                    d.addToRelatedRules(self)
    
    def apply(self, causing_cells=None):
        """
        If the first argument is done, then this method limits the digits of
        the second argument by the value of the first argument.
        
        If the second argument is done then this method sets the first
        argument to the correct value.
        
        """
        # If the second number is done then set the first numbers' value.
        if self.arg2.isDone():
            cross_sum = self.crossSum(self.arg2.value())
            if self.arg1.intersect(set([cross_sum])):
                return [self.arg1]
            else:
                return []
        
        # Otherwise, if the first argument is done...
        if not self.arg1.isDone():
            # (cannot do anthing if both arguments are not done...)
            return []
        
        result = set()
        
        # kick out "too large" and "too small" digit values
        for f, comp in ((min, op.le), (max, op.ge)):
            for digit in self.arg2.digits:
                cross_sum = self.arg1.value()
                for other in [d for d in self.arg2.digits if d != digit]:
                    if len(other.getFeasibleValues()) > 0:
                        cross_sum -= f(other.getFeasibleValues())
                
                new = set(i for i in digit.getFeasibleValues() \
                    if comp(i, cross_sum))
                if digit.force(new):
                    result.add(digit)

        def check(n):
            return self.crossSum(n) == self.arg1.value()
        if self.applyPredicate(self.arg2, check):
            result.add(self.arg2)

        if self.arg2.length == 4 and self.arg2.cardinality() > sc.MAX_EXPLICIT_NUMBERS:
            # If the cell fulfulls these requirements, explicitly listing all
            # possible numbers might be a good idea.
            self.arg2.intersect(set(filter(check, range(1000,9999))))
        
        return list(result)
    
    def crossSum(self, n):
        """
        Calculates the cross sum of value n.
        
        """
        return reduce(lambda x,y: x+y, su.digits(n), 0)
    
    def detailedDescription(self):
        return "The cross sum of %s is %s." % (str(self.arg2), str(self.arg1))


class CrossProductRule(BinaryRule):
    """
    Binary rule, the cross product of the second argument must be equal to the
    first argument's value.
    
    """
    
    def __init__(self, value, number):
        """Initialize the rule with the given cross sum value and number."""
        BinaryRule.__init__(self, value, number)
        self.name = "CrossProdIs"
    
    def register(self):
        """Register itself with arguments and their digits (if any)."""
        BinaryRule.register(self)
        
        for arg in [self.arg1, self.arg2]:
            if arg.hasDigits():
                for d in arg.digits:
                    d.addToRelatedRules(self)
    
    def apply(self, causing_cells=None):
        """
        If the first argument is done, then this method limits the digits of
        the second argument by the value of the first argument.
        
        If the second argument is done then this method sets the first
        argument to the correct value.
        
        """
        # If the second number is done then set the first numbers' value.
        if self.arg2.isDone():
            cross_prod = self.crossProd(self.arg2.value())
            if self.arg1.intersect(set([cross_prod])):
                return [self.arg1]
            else:
                return []
        
        # Otherwise, if the first argument is done...
        if not self.arg1.isDone():
            # (cannot do anthing if both arguments are not done...)
            return []
        
        result = set()
        
        # kick out digits that are no divisor of the valu
        
        # kick out "too large" and "too small" digit values
        
        # try:
        #     # kick out possible values for arg2's numbers if their cross sum
        #     # is wrong
        #     possible_values = self.arg2.getFeasibleValues()
        #     correct_values = set(n for n in possible_values \
        #         if self.crossSum(n) == self.arg1.value())
        #     if self.arg2.intersect(correct_values):
        #         result.add(self.arg2)
        # except sg.SetTooLargeException, e:
        #     # self.arg2.getFeasibleValues() threw this, swallow error
        #     pass
        
        def check(n):
            return self.crossProd(n) == self.arg1.value()
        if self.applyPredicate(self.arg2, check):
            result.add(self.arg2)

        return list(result)
    
    def crossProd(self, n):
        """
        Calculates the cross prod of value n.
        
        """
        return reduce(lambda x,y: x*y, su.digits(n), 1)

    def detailedDescription(self):
        return "The cross product of %s is %s." % (str(self.arg2), str(self.arg1))


class MultipleOfRule(BinaryRule):
    """Binary Rule, the first argument must be a multiple of the second."""
    
    def __init__(self, arg1, arg2):
        """Initialize the MultipleOfRule with the two arguments."""
        assert isinstance(arg1, sg.Number), "argument must be Number"
        assert isinstance(arg2, sg.Number), "argument must be Number"
        
        BinaryRule.__init__(self, arg1, arg2)
        self.name = 'V'
    
    def apply(self, causing_cells=None):
        """
        If the second argument is done and the first argument has digits (and
        thus the maximum number of the first argument is known) and is not 
        done yet then the first argument's values are set to all multiples of
        the second argument's values.  Also set the digits possible values
        to the values ocurring in these values.
        
        If both arguments are done then check whether the first argument if a
        multiple of the second argument.  If this is not the case then set
        the possible values of both to the empty set and return a list 
        containing both arguments.
        
        Otherwise, intersect the last digit of the first argument with all
        multiples of the second.
        
        """
        result = set()
        
        if self.arg1.isDone() and self.arg2.isDone():
            # case: both arguments are done
            #
            # return if we reached this branch since we already have done
            # digits and can only set their possible values to []!
            if self.arg1.value() % self.arg2.value() != 0:
                self.arg1.force(set())
                self.arg2.force(set())
                
                return [self.arg1, self.arg2]
            else:
                # everything is OK
                return []
        elif not self.arg1.isDone() and self.arg2.isDone():
            # case: second argument is done, first one is not
            if self.arg2.value() == 0:
                # someone was mean, it seems
                if self.arg1.intersect(set((0,))):
                    result.add(self.arg1)

            elif self.arg1.hasDigits():
                mini = self.arg1.minimum() / self.arg2.value()
                maxi = self.arg1.maximum() / self.arg2.value()
                
                # care about negative values
                if mini > maxi: mini, maxi = maxi, mini
                
                if maxi - mini < sc.MAX_EXPLICIT_NUMBERS:
                    values = set(i*self.arg2.value() \
                        for i in range(mini, maxi+1) \
                        if self.arg1.minimum() <= i*self.arg2.value() <= self.arg1.maximum())
                    if self.arg1.intersect(values):
                        result.add(self.arg1)
        
        if self.arg1.hasDigits() and self.arg2.hasDigits():
            # intersect the last digit of the first argument with all
            # multiples of the last digit of the second argument
            multiples = set((i * j) % 10 \
                for i in self.arg2.digits[-1].getFeasibleValues() \
                for j in range(0, 10))
            if self.arg1.digits[-1].intersect(multiples):
                result.add(self.arg1.digits[-1])
            
            # intersect the last digit of the second argument with all factors
            # of the last digit of the first argument modulo 10
            factors = set()
            for i in self.arg1.digits[-1].getFeasibleValues():
                factors |= su.factors_of(i)
            if self.arg2.digits[-1].intersect(factors):
                result.add(self.arg2.digits[-1])
        
        return list(result)
    
    def __str__(self):
        """Returns a human readable version of the Rule."""
        s = str(self.arg1)
        t = str(self.arg2)
        return "%s(%s, %s)" % (self.name, s, t)

    def detailedDescription(self):
        return "%s is a multiple of %s." % (str(self.arg1), str(self.arg2))


class EqualityRule(BinaryRule):
    """Binary Rule, both arguments must be equal."""
    
    def __init__(self, c1, c2):
        BinaryRule.__init__(self, c1, c2)
        self.name = '='
        self.applied = False
        
        # If this rule is applied on Digits then it should be get a higher
        # priority.
        if isinstance(self.arg1, sg.Digit):
           self.priority = PRIORITY_DIGIT_RULES
    
    def normalize(self):
        """
        If both arguments are Number objects, then create an EqualityRule
        their digits.  Return a list of these rules.
        
        Otherwise, do nothing and return an empty list.
        
        """
        if (isinstance(self.arg1,sg.Number) & isinstance(self.arg2,sg.Number) and
                self.arg1.hasDigits() & self.arg2.hasDigits()):
            
            assert len(self.arg1.digits) == len(self.arg2.digits)
            
            result = []
            
            # create rules
            for i in range(len(self.arg1.digits)):
                rule = EqualityRule(self.arg1.digits[i], self.arg2.digits[i])
                result.append(rule)
            
            return result
        else:
            return [self]
    
    def register(self):
        """
        Only register if not both arguments are Number objects.
        
        """
        if (isinstance(self.arg1,sg.Number) & isinstance(self.arg2,sg.Number) and
                self.arg1.hasDigits() & self.arg2.hasDigits()):
            return
        else:
            return BinaryRule.register(self)
    
    def apply(self, causing_cells=None):
        """
        Applies the EqualityRule.  The effect depends on the types of
        the arguments:
        
        * Number, Fixed -- The Number's Digits are set according to the Fixed
                           value
        * Number, Number -- The rule adds EqualityRules on the Numbers' 
                            Digits.
        * Digit, Digit -- Both Digits' set of valid values are set to the
                          intersection of both.
        
        """
        # If the rule has already been aplied and does not have to be applied
        # any more.
        if self.applied:
            return []
        
        # shortcuts
        arg1, arg2 = self.arg1, self.arg2
        
        if isinstance(arg1, sg.Digit):
            # if first is digit, second must be, too
            assert isinstance(arg2, sg.Digit)
            return self.applyOnDigits(arg1, arg2)
        elif isinstance(arg1, sg.Number) and isinstance(arg2, sg.Number):
            # we are working on numbers, but obviously they don’t have Digits
            return self.applyOnNumberAndNumber(arg1, arg2)
        else:
            assert False, "Combination not allowed: (%s, %s)!" % \
                    (type(arg1), type(arg2))
    
    def applyOnNumberAndNumber(self, n1, n2):
        """
        Applies rules on two numbers.
        
        If both carry a list of explicit values, force each to the
        intersection of these values
        """
        changed = []

        if n2.cardinality() < sc.MAX_EXPLICIT_NUMBERS:
            if n1.intersect(n2.getFeasibleValues()):
                changed.append(n1)
        if n1.cardinality() < sc.MAX_EXPLICIT_NUMBERS:
            if n2.intersect(n1.getFeasibleValues()):
                changed.append(n2)
        
        return changed
    
    def applyOnDigits(self, digit1, digit2):
        """Apply the equality rule on digits.  Set the possible values of
        both to the intersection of their currently possible values.
        
        """
        if digit1.getFeasibleValues() == digit2.getFeasibleValues():
            return []
        
        intersection = digit1.getFeasibleValues() & digit2.getFeasibleValues()
        
        result = []
        if digit1.getFeasibleValues() != intersection:
            digit1.force(intersection)
            result.append(digit1)
        if digit2.getFeasibleValues() != intersection:
            digit2.force(intersection)
            result.append(digit2)
        
        return result
    
    def detailedDescription(self):
        return "%s equals %s." % (str(self.arg1), str(self.arg2))


class OperatorRule(Rule):
    """Operator-As-Rule baseclass"""
    
    def __init__(self, *cells):
        Rule.__init__(self)
        self.name = 'UNKNOWN'
        self.cells = cells
        self.register()
        
        self.priority = PRIORITY_OPERATOR_RULE
    
    def register(self):
        """Register with dependencies. Override in subclass if necessary."""
        for cell in self.cells:
            cell.addToRelatedRules(self)

    def choose_each(self, list):
        """ Helper function: Returns a list of pairs of elements and the other values """
        ret = []
        for i in range(len(list)):
            ret.append( (i, list[:i]+list[i+1:]) )
        return ret
    
    def sourceCard(self):
        c = 1
        for cell in self.cells[1:]:
            c *= cell.cardinality()
        return c
    
    def apply(self, causing_cells=None):
        changed = []
        
        for target, sources in self.choose_each(self.cells):
            if causing_cells == None or reduce(op.or_, map(lambda c: c in causing_cells, sources)):
                card = reduce(op.mul, map(lambda c: c.cardinality(), sources))
                if card < sc.MAX_EXPLICIT_NUMBERS:
                    results = set()
                    for args in su.xcombine(map(lambda c: c.getFeasibleValues(), sources)):
                        res = self.calc[target](*args)
                        if res is not None:
                            res = results.add(res)
                    if self.cells[target].intersect(results):
                        changed.append(self.cells[target])

        # Try again for changed sources
        for target, sources in self.choose_each(self.cells):
            if reduce(op.or_, map(lambda c: c in changed, sources)): 
                card = reduce(op.mul, map(lambda c: c.cardinality(), sources))
                if card < sc.MAX_EXPLICIT_NUMBERS:
                    results = set()
                    for args in su.xcombine(map(lambda c: c.getFeasibleValues(), sources)):
                        res = self.calc[target](*args)
                        if res is not None:
                            res = results.add(res)
                    if self.cells[target].intersect(results):
                        changed.append(self.cells[target])

        return changed
    
    def __str__(self):
        """Returns a human readable string representation of the rule."""
        return "%s(%s)" % (self.name, ",".join(map(str, self.cells)))

    def relatedCells(self):
        return self.cells


class MultOperatorRule(OperatorRule):
    def __init__(self, *cells):
        OperatorRule.__init__(self, *cells)
        self.name = "IsMal"
    
    def execute(self, a1, a2):
        return a1 * a2

    def normalize(self):
        rules = [self]

        rules.extend(MultipleOfRule(self.cells[0], self.cells[1]).normalize())
        rules.extend(MultipleOfRule(self.cells[0], self.cells[2]).normalize())
        return rules

    def div(a,b):
        if a % b == 0:
            return a / b
        else:
            return None

    calc = (
            lambda b, c: b * c,
            div,
            div
            )
        
    def detailedDescription(self):
        return "%s is the product of %s and %s" % (str(self.cells[0]), str(self.cells[1]), str(self.cells[2]))


class AddOperatorRule(OperatorRule):
    def __init__(self, *cells):
        OperatorRule.__init__(self, *cells)
        self.name = "IsPlus"

    calc = (
            lambda b, c: b + c,
            lambda a, c: a - c,
            lambda a, b: a - b
            )
        
    def detailedDescription(self):
        return "%s is the sum of %s and %s" % (str(self.cells[0]), str(self.cells[1]), str(self.cells[2]))


class SubstractOperatorRule(OperatorRule):
    def __init__(self, *cells):
        OperatorRule.__init__(self, *cells)
        self.name = "IsMinus"
    
    calc = (
            lambda b, c: b - c,
            lambda a, c: a + c,
            lambda a, b: b - a
            )

    def detailedDescription(self):
        return "%s is the difference of %s and %s" % (str(self.cells[0]), str(self.cells[1]), str(self.cells[2]))


class FlipOperatorRule(BinaryRule):
    def __init__(self, *cells):
        BinaryRule.__init__(self, *cells)
        self.name = "IsReverseOf"

    def copyDigits(self, fromNumber, toNumber):
        assert toNumber.digits == None or toNumber.digits[0] == None
        toNumber.digits = fromNumber.digits[::-1]
        toNumber.length = fromNumber.length

    def normalize(self):
        if (isinstance(self.arg1, sg.FieldNumber)
                and isinstance(self.arg2, sg.FieldNumber) 
                and len(self.arg1.digits) == len(self.arg2.digits)):
            # Both cells have digits, and the same number of them
            # so generate Equality-Rules

            new_rules = []
            
            for i in range(len(self.arg1.digits)):
                rule = EqualityRule(self.arg1.digits[i], self.arg2.digits[-(i+1)]).normalize()
                new_rules.extend(rule)
            
            return new_rules
        else:
            return [self]

    def register(self):
        # Logic already happend in normalize() or __init__()
        if (isinstance(self.arg1, sg.FieldNumber)
                and isinstance(self.arg2, sg.FieldNumber) 
                and len(self.arg1.digits) == len(self.arg2.digits)):
            pass
        else:
            BinaryRule.register(self)

    def apply(self, causing_cells=None):
        changed = []
        
        # right to left
        if self.arg2.cardinality() < sc.MAX_EXPLICIT_NUMBERS:
            results = set()
            for arg in self.arg2.getFeasibleValues():
                results.add(self.execute(arg))
            if self.arg1.intersect(results):
                    changed.append(self.arg1)

        # left to right, pad if nessary
        if self.arg1.cardinality() < sc.MAX_EXPLICIT_NUMBERS and self.arg2.length:
            results = set()
            for arg in self.arg1.getFeasibleValues():
                digits = su.digits(arg)[::-1]
                zeros  = [0]*(self.arg2.length - len(digits))
                r = su.assembleDigits(su.digits(arg)[::-1] + zeros)
                results.add(r)
            if self.arg2.intersect(results):
                    changed.append(self.arg2)

        # try again right to left, who knows
        if self.arg2.cardinality() < sc.MAX_EXPLICIT_NUMBERS:
            results = set()
            for arg in self.arg2.getFeasibleValues():
                results.add(self.execute(arg))
            if self.arg1.intersect(results):
                    changed.append(self.arg1)

        return changed
    
    
    def execute(self, arg):
        return su.assembleDigits(su.digits(arg)[::-1])
        
    def detailedDescription(self):
        return "%s is %s reversed" % (str(self.cells[0]), str(self.cells[1]))


class NumberRule(Rule):
    """
    Special Rule per Number. Tasks:
    
    * Propagates Changes in the Digits to the Number
    
    Later:
    
    * Once the Number also maintains a list of valid options, keep that 
      two-way in sync with with the digits
    
    """
    def __init__(self, number):
        Rule.__init__(self)
        self.number = number
        self.name = "NR"
        self.last_card = None
        
        self.priority = PRIORITY_NUMBER_RULE
        
        number.addToRelatedRules(self)
        for digit in number.digits:
            digit.addToRelatedRules(self)
    
    def apply(self, causing_cells=None):
        changed = []
        
        # remove numbers unless allowed by digits
        
        # only apply if the feasible values have been set
        # TODO: encapsulate access to _feasible_values so getting them
        # does not possibly create a too large set
        if self.number._feasible_values != None:
            # invalidate numbers that contain already invalid digits
            # only run if at least one digit was changed
            if causing_cells is None or filter(lambda d: d in causing_cells, self.number.digits):
                invalid = set()
                for val in self.number.getFeasibleValues():
                    digits = su.digits(val,len(self.number.digits))
                    if digits is None:
                        invalid.add(val)
                    else:
                        for d1, d2 in zip(digits, self.number.digits):
                            if d1 not in d2.getFeasibleValues():
                                invalid.add(val)
                                break
                if invalid:
                    self.number.invalidateSet(invalid)
                    changed.append(self.number)
            
            # remove digits unless allowed by numbers
            # only run when digit has changed
            if self.number in changed or causing_cells is None or self.number in causing_cells:
                allowed = [set() for i in range(self.number.length)]
                for val in self.number.getFeasibleValues():
                    digits = su.digits(val, self.number.length)
                    for d, a in  zip(digits, allowed):
                        a.add(d)
                
                for a, d in zip(allowed, self.number.digits):
                    if d.intersect(a):
                        changed.append(d)
        else:
            if self.number.cardinality() < sc.MAX_EXPLICIT_NUMBERS:
                self.number.getFeasibleValues()
                changed.append(self.number)
        
        return changed
    
    def __str__(self):
        return "NumberRule for %s" % str(self.number)

    def relatedCells(self):
        return (self.number,)
        
    def detailedDescription(self):
        return "Synchronize knowledge about the possible digits and the possible values of %s" % (str(self.number))


class CrossSumUniqueCombinationRule(BinaryRule):
    """
    This rule is the special case if both the cross sum of a number and the
    fact that the number's digits are unique are known.
    
    TODO: Implement this!
    TODO: Test this!
    
    """
    
    def __init__(self, value, number):
        BinaryRule.__init__(self, value, number)
    
    def register(self):
        """This rule does not have to register itself, it only runs once."""
        return
    
    def apply(self):
        """
        Based on the value and number of digits of the number, we can 
        efficiently compute all possible values of the number.
        
        """
        # TODO: Implement me.
        # def gen(n, r):
        #             """n -- sum to reach, r -- length of number"""
        #             def gen_helper(n, r, s, l):
        #                 """
        #                 Actual implementation of gen().
        #                 
        #                 n -- sum left
        #                 r -- number len left
        #                 s -- sum so far
        #                 l -- list so far
        #                 
        #                 """
        #                 # if we would return 0 in the leaf anyway
        #                 if n < 0 or n - r * 9 > 0 or n - r + 1 < 0:
        #                     return []
        #                 
        #                 if r == 0:
        #                     # return the end of the tree's path if n < 10 (limitation
        #                     # because the entries of l are the number's places)
        #                     if n < 10 and n not in l:
        #                         return [l + [n]]
        #                     else:
        #                         return []
        #                 else:
        #                     # first place must be > 0
        #                     my_min = 0 if l else 1
        #                     # generate next possible addends
        #                     res = []
        #                     for i in range(my_min, min(n + 1, 10)):
        #                         if i in l: continue # skip dupes
        #                         res += gen_helper(n - i, r - 1, s + i, l + [i])
        #                     return res
        #             
        #             return gen_helper(n, r, 0, [])
        

NAME_TO_RULE = {"AB": DescendingRule, "K":CubicRule,
    "PAL":PalindromeRule, "PZ":PrimeRule, "Q":SquareRule, 
    "U":UniqueDigitsRule, "V":MultipleOfRule, "=":EqualityRule}

OPERATOR_TO_RULE = {'mal': MultOperatorRule, 
    'minus': SubstractOperatorRule, 'plus': AddOperatorRule,
    'qp': CrossProductRule, 'qs': CrossSumRule,
    'r': FlipOperatorRule}

# vim:sw=4:ts=4:sts=4:expandtab

