#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


"""
This module contains the FieldWidget that is used to draw sevendown.graph.Field
objects.

"""

import gtk.glade
import gtk.gdk
import gtk
import gobject
import cairo
import pygtk
import math

import sevendown.loader as sl
import sevendown.solver as ss
import sevendown.logger
import sevendown.graph
import sevendown.pretty
import sevendown.undo

class PaintingInfo(object):
    """
    Stores information required by the FieldPainter computed from the
    DrawingArea's size and Field's.
    
    Attributes:
    
    * cellSize
    * padX, padY
    
    """
    
    def __init__(self, field, width, height):
        cells_x, cells_y = field.width, field.height
        
        self.cellSize = min((width-10)/cells_x, (height-10)/cells_y)
        self.padX = (width - cells_x * self.cellSize) / 2
        self.padY = (height - cells_y * self.cellSize) / 2
    

class FieldPainter(object):
    # cr -- Cairo Rendering context
    
    def draw(self, cr, rect, field, highlights, splits_to_try):
        
        if field == None:
            field = sevendown.graph.Field(8,8)
            info = PaintingInfo(field, rect.width, rect.height)
            self.drawFieldWhenEmpty(cr, info, field, rect)
            return # don't draw
        
        info = PaintingInfo(field, rect.width, rect.height)
        self.drawBackground(cr, info)
        self.drawHighlights(cr, info, field, highlights)
        self.drawGrid(cr, info, field)
        self.drawMarkers(cr, info, field)
        self.drawDigits(cr, info, field, splits_to_try)
        
        # self.drawClickableAreaOverlay(cr, info, field)
    
    def drawFieldWhenEmpty(self, cr, info, field, rect):
        """Draws a dummyField with fieldSize = 8, if nothing else is loaded."""
        cr.set_source_rgb(1, 1, 1)
        cr.paint()
        self.drawGrid(cr, info, field)
        for x in range(0, 8):
            for y in range(0, 8):
                for digit in range(0, 10):
                    self.drawDigit(cr, info, x, y, digit, (0, 0, 0))
        
        width, height = rect.width, rect.height
        self.drawLogo(cr, width, height, 2)

    def drawLogo(self, cr, width, height, div):
        # Something logolike
        cr.set_font_size(height / div)
        x_b, y_b, w, h, _, _ =  cr.text_extents('7')
        cr.set_font_size((height * 2)/ (div * 7))
        x_b2, y_b2, w2, h2, _, _ =  cr.text_extents('down')
        
        cr.set_line_join(cairo.LINE_JOIN_ROUND)
        cr.rectangle((width - w -10)/2 -  h2, (height-h-10)/2,
                     (        w +10)   +2*h2, (       h+10)  )
        cr.set_source_rgba(1,1,1,0.8)
        cr.fill_preserve()
        cr.set_source_rgba(0,0,0,0.3)
        cr.set_line_width(4)
        cr.stroke()
        
        cr.set_font_size(height / div)
        cr.move_to((width - w)/2 - x_b, (height-h)/2 - y_b)
        cr.set_source_rgb(1,0,0)
        cr.show_text('7')
        
        cr.set_source_rgb(0,0,0)
        cr.move_to((width + w)/2, (height-h)/2)
        cr.set_font_size((height * 2)/ (div * 7))
        cr.rotate(math.pi / 2)
        cr.show_text('down')
    
    def drawBackground(self, cr, info):
        """Draws the background."""
        cr.set_source_rgb(1, 1, 1)
        cr.paint()
    
    def drawHighlights(self, cr, info, field, highlights):
        # Highlights
        for cell, color in highlights.iteritems():
            if isinstance(cell, sevendown.graph.Number) and cell.hasDigits():
                for digit in cell.digits:
                    if digit.coords:
                        self.drawDigitHighlight(cr, info, digit.coords, color)
            elif isinstance(cell, sevendown.graph.Digit):
                if cell.coords:
                    self.drawDigitHighlight(cr, info, cell.coords, color)
    
    def drawDigitHighlight(self, cr, info, (x,y), color):
        """
        Highlights the field at the given coordinate, with a given color.
        
        """
        cr.rectangle(info.padX + x * info.cellSize,
                     info.padY + y * info.cellSize,
                     info.cellSize, info.cellSize)
        cr.set_source_rgb(*color)
        cr.fill()
    
    def drawGrid(self, cr, info, field):
        """Draws the bulk field, which means: background and grid."""
        # drawing normal grid
        cr.translate(0.5, 0.5) # To avoid aliasing on odd line widths
        cr.set_line_width(1)
        cr.set_line_cap(cairo.LINE_CAP_BUTT)
        cr.set_line_join(cairo.LINE_JOIN_MITER)
        cr.set_source_rgb(0, 0, 0)
        for i in range(field.width + 1):
            x = info.padX + i * info.cellSize
            cr.move_to(x, info.padY)
            cr.rel_line_to(0, field.height * info.cellSize)
        
        for i in range(field.height + 1):
            y = info.padY + i * info.cellSize
            cr.move_to(info.padX, y)
            cr.rel_line_to(field.width * info.cellSize, 0)
        cr.stroke()
        
        # draw thick lines
        cr.set_line_width(3)
        # frame
        cr.set_line_cap(cairo.LINE_CAP_ROUND)
        cr.set_line_join(cairo.LINE_JOIN_BEVEL)
        cr.rectangle(info.padX, info.padY,
                     field.width * info.cellSize, field.height * info.cellSize)
        cr.stroke()
        
        # intermediate borders
        cr.set_line_cap(cairo.LINE_CAP_SQUARE)
        cr.set_line_join(cairo.LINE_JOIN_MITER)
        if field:
            for row in range(field.height):
                for col in range(field.width - 1):
                    if sevendown.loader.BORDER_RIGHT in field.schema[row][col]:
                        cr.move_to(info.padX + (col+1) * info.cellSize,
                                   info.padY + (row)   * info.cellSize)
                        cr.rel_line_to(0, info.cellSize)
                        cr.stroke()
            for row in range(field.height - 1):
                for col in range(field.width):
                    if sevendown.loader.BORDER_BOTTOM in field.schema[row][col]:
                        cr.move_to(info.padX + (col)   * info.cellSize,
                                   info.padY + (row+1) * info.cellSize)
                        cr.rel_line_to(info.cellSize, 0)
                        cr.stroke()
    
    def drawMarkers(self, cr, info, field):
        """Draw all markers"""
        if field.markers:
            for m, x, y in field.markers.itervalues():
                self.drawMarker(cr, info, x, y, m, (0,0,0,0.2))
    
    def drawMarker(self, cr, info, xpos, ypos, marker, color):
        """Draws given marker at position (xpos, ypos) in the given color."""
        
        xLetterSize = int(info.cellSize * 6.0/8.0)
        
        cr.set_font_size(xLetterSize)
        cr.set_source_rgba(*color)
        
        x_b2, y_b2, w2, h2, _, _ =  cr.text_extents(marker)
        
        cellX = info.padX + xpos * info.cellSize 
        cellY = info.padY + (ypos + 1) * info.cellSize 
        
        cr.move_to(cellX, cellY - h2 - y_b2)
        cr.show_text(marker)
    
    def drawDigits(self, cr, info, field, splits_to_try):
        """Draw all digits."""
        # Regular digits
        for y, line in enumerate(field.digits):
            for x, digit in enumerate(line):
                for d in digit.getFeasibleValues():
                    self.drawDigit(cr, info, x, y, d, (0,0,0))
        
        # Blue digits: Splits that are yet to try
        for cell, splits in splits_to_try:
            if cell.coords:
                x,y = cell.coords
                for split in splits:
                    for d in split:
                        self.drawDigit(cr, info, x, y, d, (0,0,1))
    
    def drawDigit(self, cr, info, xpos, ypos, digit, color):
        """Draws given digit at position (xpos, ypos) in the given color."""
        digitXY = ((2,4), 
                   (1,1), (2,1), (3,1),
                   (1,2), (2,2), (3,2),
                   (1,3), (2,3), (3,3))
        
        xLetterSize = int(info.cellSize * 2.0/8.0)
        yLetterSize = int(info.cellSize * 2.0/10.0)
        yLetterSpace = int((info.cellSize * 2.0/10.0))
        
        cr.set_font_size(xLetterSize)
        cr.set_source_rgb(*color)
        
        cellX = info.padX + xpos * info.cellSize 
        cellY = info.padY + ypos * info.cellSize + (yLetterSpace / 2)
        
        cr.move_to(cellX + (digitXY[digit][0] - 0.3 )* xLetterSize,
                   cellY + digitXY[digit][1] * yLetterSize)
        cr.show_text(str(digit))
    
    def drawClickableAreaOverlay(self, cr, info, field):
        for y, line in enumerate(field.digits):
            for x, digit in enumerate(line):
                for d in range(0, 10):
                    self.drawOverlayFor(cr, info, x, y, d)
    
    def drawOverlayFor(self, cr, info, xpos, ypos, digit):
        digitXY = ((2,4),
                   (1,1), (2,1), (3,1),
                   (1,2), (2,2), (3,2),
                   (1,3), (2,3), (3,3))
        
        xLetterSize = int(info.cellSize * 2.0/8.0)
        yLetterSize = int(info.cellSize * 2.0/10.0)
        yLetterSpace = int((info.cellSize * 2.0/10.0))
        
        cr.set_source_rgb(1, 0, 0)
        cr.set_line_width(0.5)
        
        cellX = info.padX + xpos * info.cellSize 
        cellY = info.padY + ypos * info.cellSize + (yLetterSpace / 2)
        
        cr.rectangle(cellX + (digitXY[digit][0] - 0.5)* xLetterSize,
                     cellY + (digitXY[digit][1] - 1) * yLetterSize,
                     xLetterSize, yLetterSize)
        cr.stroke()
    

class FieldWidgetEventHandler(object):
    """
    Actually handles even for the FieldWidget class.
    
    """
    def __init__(self, field, widget):
        self.field = field
        self.widget = widget
    
    def do_button_event(self, widget, event):
        """The button event virtual method."""
        assert widget == self.widget
        if self.widget.field == None: return
        if event.button == 1:
            x, y = event.x, event.y
            width, height = self.widget.window.get_size()
            self.handle_toggle_digit(x, y, width, height, event)
        return True
    
    def handle_toggle_digit(self, x, y, width, height, event):
        """
        Handle toggling of digits.  The user clicked on (x, y) and the current
        widget size is (width, height).
        
        """
        if self.widget.field == None: return
        location = self.point_to_cell_and_value(x, y, width, height)
        if location == False: return
        digit_x, digit_y, value = location

        shift = bool(event.state & gtk.gdk.SHIFT_MASK)
        
        digit = self.field.digits[digit_y][digit_x]
        self.widget.emit('digit-clicked', shift, digit, value)
    
    def point_to_cell_and_value(self, x, y, width, height):
        """
        Converts the point the user clicked on (x, y) and the field height and
        width (width, height) into the field's cell index and the value that
        has been clicked.
        
        Returns (cell x, cell y, value) if the location is valid and False
        otherwise.
        
        """
        if self.widget.field == None: return
        DEBUG = False
        
        info = PaintingInfo(self.field, width, height)
        x, y = x - info.padX, y - info.padY
        cell_x, cell_y = int(x / info.cellSize), int(y / info.cellSize)
        
        # ignore if out of actual field
        if not 0 <= cell_x < self.field.width: return False
        if not 0 <= cell_y < self.field.height: return False
        
        if DEBUG: print "cell (%d, %d)" % (cell_x, cell_y)
        
        # get coordinates within cell
        x = x - cell_x * info.cellSize
        y = y - cell_y * info.cellSize
        
        if DEBUG: print "within cell: (%d, %d)" % (x, y)
        
        # see FieldPainter.drawOverlayFor(...)
        digitXY = ((2,4),
                   (1,1), (2,1), (3,1),
                   (1,2), (2,2), (3,2),
                   (1,3), (2,3), (3,3))
        
        xLetterSize = int(info.cellSize * 2.0/8.0)
        yLetterSize = int(info.cellSize * 2.0/10.0)
        yLetterSpace = int((info.cellSize * 2.0/10.0))
        
        value_x = int(x / xLetterSize + 0.5) - 1
        value_y = int((y - yLetterSpace/2) / yLetterSize + 1) - 1
        
        if not 0 <= value_x < 3: return False
        if not 0 <= value_y < 4: return False
        
        value2int_map = ((1, 2, 3), (4, 5, 6), (7, 8, 9), (-1, 0, -1))
        value = value2int_map[value_y][value_x]
        
        if value == -1: return False # ignore invalid
        
        if DEBUG: print "value = %d" % value
        
        return (cell_x, cell_y, value)
    

class FieldWidget(gtk.DrawingArea):
    __gsignals__ = {
        'digit-clicked': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
                          (gobject.TYPE_BOOLEAN, gobject.TYPE_PYOBJECT, gobject.TYPE_INT,))
        }

    def __init__(self):
        """Initialize the widget's internal state."""
        gtk.DrawingArea.__init__(self)
        self.connect("expose_event", self.expose)
        self.set_size_request(400, 400)
        self.field = None
        self.solver = None
        self.painter = FieldPainter()
        self.add_events(gtk.gdk.BUTTON_PRESS_MASK)
        self.event_handler = FieldWidgetEventHandler(self.field, self)
        self.connect("button_press_event",
                self.event_handler.do_button_event)
        self.highlights = {}
    
    def set_field_and_solver(self, field, solver):
        self.field = field
        self.event_handler.field = self.field
        self.solver = solver
    
    def expose(self, widget, event):
        context = widget.window.cairo_create()
        
        # set a clip region for the expose event
        context.rectangle(event.area.x, event.area.y,
                               event.area.width, event.area.height)
        context.clip()
        
        if self.solver != None:
            self.painter.draw(context, self.get_allocation(), self.field,
                self.highlights, self.solver.splits_to_try)
        else:
            self.painter.draw(context, self.get_allocation(), self.field,
                self.highlights, None)
            
        
        return False
    
    def setHighlights(self, highlights):
        """
        Marks a digit/number to be highlighted with a certain color.
        
        highlights is of the form: {<cell object> : (r,g,b)}.
        
        Set to the empty dict to remove all highlighting.
        
        """
        self.highlights = highlights
        self.queue_draw()
    
    def get_size_requisition(self):
        """Returns (width, height) of required size."""
        return (400, 400)
    

# vim:sw=4:ts=4:sts=4:expandtab
